using System.Threading;
using Rhizome.Contract;

namespace Rhizome.Model
{
    public partial class ActiveObjectHost
    {
        #region Nested type: Dispatcher

        private class Dispatcher
        {
            private readonly IActiveObject _first; //  
            private readonly ActiveObjectHost _host;
            private readonly IActiveObject _last; //  
            private readonly object _monitor = new object();
            private IActiveObject _current; //    
            private IActiveObject _currentPrev; //    

            public Dispatcher(ActiveObjectHost host)
            {
                _host = host;
                _first = new StackActiveObject(host);
                _last = new StackActiveObject(host);
                InitList();
            }

            public int Count { get; private set; }

            private void InitList()
            {
                _first.Next = _last;
                _last.Prev = _first;
                _current = _currentPrev = null;
                Count = 0;
            }

            public void Add(IActiveObject obj)
            {
                lock (this)
                {
                    if (obj.Next != null || obj.Exception != null || obj.Sleeping) return;

                    obj.Prev = _last.Prev;
                    obj.Next = _last;
                    obj.Prev.Next = _last.Prev = obj;
                    ++Count;

                    Pulse();
                }
            }

            public void Pulse()
            {
                lock (_monitor) Monitor.PulseAll(_monitor);
            }

            public void Remove(IActiveObject obj)
            {
                lock (this)
                {
                    if (obj.Next == null) return;

                    if (obj == _currentPrev)
                        _currentPrev = obj.Prev;

                    obj.Prev.Next = obj.Next;
                    obj.Next.Prev = obj.Prev;
                    obj.Next = obj.Prev = null;
                    --Count;
                }
            }

            public IActiveObject GetFirst()
            {
                lock (this)
                {
                    _current = _first.Next;
                    _currentPrev = null;

                    if (_current == _last) //  
                        _current = null;

                    return _current;
                }
            }

            public IActiveObject GetNext()
            {
                lock (this)
                {
                    if (_current != null)
                    {
                        _current = _current.Next ?? (_currentPrev == null ? _first.Next : _currentPrev.Next);

                        if (_current == _last) //   
                            _current = _currentPrev = null;

                        else if (_current != null)
                            _currentPrev = _current.Prev;
                    }
                    return _current;
                }
            }

            public void Clear()
            {
                lock (this)
                {
                    _current = _first.Next;
                    while (_current != _last)
                    {
                        var next = _current.Next;
                        _current.Prev = _current.Next = null;
                        _current = next;
                    }
                    InitList();
                }
            }

            public void Sleep(IActiveObject obj, int timeout)
            {
                if (obj.Exception != null) return;

                lock (this)
                {
                    if (obj.Sleeping)
                        _host._timer.Remove(obj);
                    else
                        Remove(obj);

                    obj.Timeout = timeout;
                    obj.Sleeping = true;
                }
                _host._timer.Add(obj);
            }

            public void Sleep(IActiveObject obj)
            {
                lock (this)
                {
                    if (obj.Sleeping || obj.Exception != null) return;

                    obj.Sleeping = true;
                    Remove(obj);
                }
            }

            public void WakeUpByTimer(IActiveObject obj)
            {
                lock (this)
                {
                    if (obj.Timeout == 0 || !obj.Sleeping) return;

                    obj.Timeout = 0;
                    obj.Sleeping = false;
                    Add(obj);
                }
            }

            public void WakeUp(IActiveObject obj)
            {
                lock (this)
                {
                    if (obj.Timeout != 0 || !obj.Sleeping)
                        return;

                    //obj.Timeout = 0; // ?
                    obj.Sleeping = false;
                    //host.timer.Remove(obj); // ?
                    Add(obj);
                }
            }

            public void Wait()
            {
                lock (_monitor) Monitor.Wait(_monitor, _host._timer.SleepTime);
            }
        }

        #endregion
    }
}