using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using System.Windows.Markup;
using Rhizome.Contract;
using Rhizome.Internal;
using Rhizome.Model.Validation;

namespace Rhizome.Model
{
	[ContentProperty("States")]
	public abstract class AutomatonBase<TState, TPlant> : RhizomeElement, IEventSink, IValidate, IAutomaton,
	                                                      IStateInitialize
		where TState : class, IIdentityAccessor<object>
	{
		private readonly RhizomeElementKeyedCollection<Edge> _edges;
		private readonly Dictionary<TState, object> _history = new Dictionary<TState, object>();
		private readonly object _stateLock = new object();
		private readonly RhizomeElementKeyedCollection<TState> _states;
		private bool _isInTransition;
		private TState _state;

		protected AutomatonBase()
		{
			_states = new RhizomeElementKeyedCollection<TState>(this);
			_edges = new RhizomeElementKeyedCollection<Edge>(this);
		}

		public KeyedCollection<object, TState> States
		{
			[DebuggerStepThrough]
			get { return _states; }
		}

		public KeyedCollection<object, Edge> Edges
		{
			[DebuggerStepThrough]
			get { return _edges; }
		}

		/// <summary>
		/// 	Initial state ID
		/// </summary>
		public object InitialState { get; set; }

		public TState State
		{
			[DebuggerStepThrough]
			get { return _state; }
			private set
			{
				if (_state == value) return;
				if (Logger != null) Logger.WriteLine("{0}.set_State({1})", this, value);
				_state = value;
			}
		}

		public TPlant Plant
		{
			[DebuggerStepThrough]
			get { return (TPlant) GetValue(PlantProperty); }
			[DebuggerStepThrough]
			set { SetValue(PlantProperty, value); }
		}

		#region Overrides of RhizomeElement

		protected override IEnumerable<IEnumerable> GetElementCollections()
		{
			return base.GetElementCollections().Concat(new IEnumerable[] {Edges, States});
		}

		#endregion

		#region Implementation of IAutomaton

		bool IAutomaton.ContainsState(object stateId)
		{
			return States.Contains(stateId);
		}

		#endregion

		#region IEventSink Members

		public virtual void CastEvent(object source, object @event)
		{
			if (source == null) throw new ArgumentNullException("source");
			if (@event == null) throw new ArgumentNullException("event");

			if (Monitor.TryEnter(_stateLock) == false)
			{
				if (Logger != null)
					Logger.WriteLine(
						"(sic!) {0}.CastEvent(Source = {1}, Event = {2}) is locked by another thread",
						this, source, @event);
				return;
			}

			try
			{
				if (Logger != null) Logger.WriteLine("{0}.CastEvent(Source = {1}, Event = {2})", this, source, @event);

				if (Equals(source, State) == false)
				{
					if (Logger != null) Logger.WriteLine("(sic!) {0}.CastEvent. State != Source ({1})", this, source);
					return;
				}

				if (_isInTransition)
				{
					throw new InvalidOperationException(
						string.Format(
							"{0}.CastEvent(Source = {1}, Event = {2}) fired in transition state. Use IForwardEvent instead",
							this, source, @event));
				}


				_isInTransition = true;
				_history.Clear();

				var stateInitialize = State as IStateInitialize;
				if (stateInitialize != null) stateInitialize.Cleanup();

				InternalCastEvent(State, @event);
			}
			finally
			{
				_isInTransition = false;
				Monitor.Exit(_stateLock);
			}
		}

		#endregion

		#region IValidate Members

		public virtual IEnumerable<IValidationResult> GetValidationResults()
		{
			if (InitialState == null)
				yield return ValidationResult.InitialStateNotSet;

			else if (States.Contains(InitialState) == false)
				yield return ValidationResult.StateNotFound(InitialState);
		}

		#endregion

		private void InternalCastEvent(TState fromState, object @event)
		{
			var stateId = fromState.Id;
			if (Edges.Contains(stateId) == false)
				throw new EdgeNotFoundException<TState>(State, @event);

			var edge = Edges[stateId];
			if (edge.Transitions.Contains(@event) == false)
				throw new EdgeNotFoundException<TState>(State, @event);

			var transition = edge.Transitions[@event];
			var nextStateId = transition.GetTransitionState();
			if (States.Contains(nextStateId) == false)
				throw new StateNotFoundException(stateId);

			// exiting state
			var stateActions = fromState as IStateActions;
			if (stateActions != null) stateActions.ExitAction();

			// edge transition action
			var actionSource = transition as IActionSource;
			var transitionAction = actionSource == null ? null : actionSource.Action;
			if (transitionAction != null) transitionAction.Process();

			ChangeState(fromState, nextStateId);
		}

		private void ChangeState(TState fromState, object nextStateId)
		{
			if (nextStateId == null)
			{
				OnStop();
				return;
			}

			var nextState = States[nextStateId];

			var stateActions = nextState as IStateActions;
			if (stateActions != null) stateActions.EntryAction();

			var forwardEvent = nextState as IForwardEvent;
			if (forwardEvent != null)
			{
				var e = forwardEvent.GetForwardEvent();
				if (e != null)
				{
					if (Logger != null)
						Logger.WriteLine(
							"{0}[{1}].InternalCastEvent(NextState = {2}, ForwardEvent = {3})",
							GetType().Name, fromState, nextState, e);

					if (fromState != null)
					{
						if (_history.ContainsKey(fromState))
						{
							throw new InvalidOperationException(
								"Event Sink infinite loop detected: " +
								FormatHistory());
						}

						_history.Add(fromState, e);
					}

					// todo: remove recursion?
					InternalCastEvent(nextState, e);
					return;
				}
			}

			// set new state
			State = nextState;

			var init = nextState as IStateInitialize;
			if (init != null) init.Initialize();
		}

		private string FormatHistory()
		{
			return (from p in _history select "{0} + {1}".With(p.Key, p.Value)).ToArray().Join(" -> ");
		}

		public override T GetValue<T>(RhizomeProperty<T> property)
		{
			if (ReferenceEquals(property, EventSyncProperty) || ReferenceEquals(property, AutomatonProperty))
				return (T) ((object) this);

			return base.GetValue(property);
		}

		public void Start()
		{
			if (State != null) throw new InvalidOperationException("State must be null");
			OnStart();
		}

		public void Stop()
		{
			if (State == null) return;
			OnStop();
		}

		protected virtual void OnStart()
		{
			ChangeState(null, InitialState);
		}

		protected virtual void OnStop()
		{
			var stateInitialize = State as IStateInitialize;
			if (stateInitialize != null) stateInitialize.Cleanup();
			State = null;
		}

		public override string ToString()
		{
			return string.Format("{0}[{1}]", GetType().Name, State == null ? "null" : State.Id);
		}

		#region Implementation of IStateInitialize

		void IStateInitialize.Initialize()
		{
			OnStart();
		}

		void IStateInitialize.Cleanup()
		{
			OnStop();
		}

		#endregion
	}
}