﻿using System;
using System.Diagnostics;
using System.Threading;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Rhizome.Contract;
using Rhizome.Model;
using Rhizome.Model.Validation;

namespace Rhizome.Core.Test
{
	[TestClass]
	public class AutomatonTest
	{
		[TestMethod]
		public void EnsuareEventCanBeCastedOncePerState()
		{
			var a = new Automaton();
			a.InitialState = a.StateA.Id;
			a.Validate();
			a.Start();

			for (var i = 0; i < 500; i++)
			{
				a.Process();
				//Thread.Sleep(10);
				Assert.AreEqual(a.StateB, a.State, "Wrong state");
				Debug.WriteLine("***");

                a.Process();
                //Thread.Sleep(10);
                Assert.AreEqual(a.StateA, a.State, "Wrong sate");
                Debug.WriteLine("***");
			}
		}
	}

	internal abstract class StateBase : StateBase<object>
	{
		public static readonly object EventA = "Event_A";
		public static readonly object EventB = "Event_B";
		public static readonly object EventC = "Event_C";

		protected readonly object Event1;
		protected readonly object Event2;

		protected StateBase(object event1, object event2)
		{
			Event1 = event1;
			Event2 = event2;
		}

		public virtual void Process()
		{
			Debug.WriteLine("{0}.Process()", this);
		}

	  
	}

    internal sealed class StateA : StateBase, IStateInitialize
	{
		private Thread _t2;

		public StateA() : base(EventB, EventC)
		{
		}

		public override void Process()
		{
			base.Process();

			lock (this)
			{
				_t2 = new Thread(
					() =>
						{
							Thread.Sleep(TimeSpan.FromMilliseconds(20));
							EventSink.CastEvent(this, Event2);
						}) {Name = "2", Priority = ThreadPriority.BelowNormal};

				_t2.Start();
			}
			EventSink.CastEvent(this, Event1);
		}

        void IStateInitialize.Initialize()
	    {
	        
	    }

        void IStateInitialize.Cleanup()
		{
			Debug.WriteLine("{0}.ExitAction()", this);

			lock (this)
			{
				if (_t2 != Thread.CurrentThread) _t2.Join();
				_t2 = null;
			}
		}
	}

	internal sealed class StateB : StateBase, IForwardEvent
	{
		public StateB() : base(EventA, EventC)
		{
		}

		public override void Process()
		{
			base.Process();

            EventSink.CastEvent(this, Event1);
			EventSink.CastEvent(this, Event2);
		}

	    object IForwardEvent.GetForwardEvent()
	    {
	        return null;
	    }
	}

    internal sealed class StateC : StateBase, IStateInitialize
	{
		public StateC() : base(EventA, EventB)
		{
		}

        void IStateInitialize.Initialize()
		{
			Assert.Fail("State C is not in execution path");
		}

        public void Cleanup()
        {
            
        }
	}

	internal class Automaton : AutomatonBase<StateBase, object>
	{
		public readonly StateBase StateA = new StateA {Id = "State_A"};
		public readonly StateBase StateB = new StateB {Id = "State_B"};
		public readonly StateBase StateC = new StateC {Id = "State_C"};

		public Automaton()
		{
			States.Add(StateA);
			States.Add(StateB);
			States.Add(StateC);

			var edgeA = new Edge {State = StateA.Id};
			edgeA.Transitions.Add(new Transition {Event = StateBase.EventB, State = StateB.Id});
			edgeA.Transitions.Add(new Transition {Event = StateBase.EventC, State = StateC.Id});
			Edges.Add(edgeA);

			var edgeB = new Edge {State = StateB.Id};
			edgeB.Transitions.Add(new Transition {Event = StateBase.EventA, State = StateA.Id});
			edgeB.Transitions.Add(new Transition {Event = StateBase.EventC, State = StateC.Id});
			Edges.Add(edgeB);
		}

		public void Process()
		{
			State.Process();
		}
	}
}